`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: Steven Howell
// 
// Create Date: 07/17/2017 05:56:35 PM
// Design Name: 
// Module Name: navigation
// Project Name: 
// Target Devices: XC7A35TCSG324-1L
// Tool Versions: Vivado 2017.1
// Description: Navigation module, Whippersnapper Runt robot
// 
// Dependencies: 
// 
// Revision:
// Revision 0.01 - File Created
// Additional Comments:
// 
//////////////////////////////////////////////////////////////////////////////////


module navigation (input raw_clk_p, input rstn,
    // I2C interface
    inout SDA, inout SCL,
    // User/HMC5883L signals
    input I2C_RST, input INIT, output data_out_led, output ack_or_complete, input drdy,
    input cal_complete,
    
    // commanded magnetic heading from top level, forward and reverse signals
    input cmd_hdg_sel, input fwd, input rev,
    
    // motor control and position encoder I/O
    output A_1A, output A_1B, /* left front motor */ output A_2A, output A_2B, /* left rear motor */
    output B_1A, output B_1B, /* right front motor */ output B_2A, output B_2B, /* right rear motor */
    
    // uart outputs used for debug
    output uart_tx, input uart_rx,
    
    // output reset_n to hold TI PHY in powerdown
    output eth_phy_rstn
    
    );
    
    wire clk_slow;          // the slowest clock around 200 kHz for I2C comms, motor control, etc.
                            // comes from clock module in the mag compass block, goes to motor control
    wire clk_fast;          // the fast clock at 100MHz, used for debug uart
    
    wire A_1A_o, A_1B_o, A_2A_o, A_2B_o, B_1A_o, B_1B_o, B_2A_o, B_2B_o;
                            
    wire [23:0] mag_angle;  // main output of top_HC5883L compass module.  Radian value, 14 bits fractional
    wire mag_angle_valid;
    reg [23:0] cmd_hdg;
    
   assign eth_phy_rstn = 1'b0;      // Keep the ethernet PHY on board always in reset to reduce ~267 mW power
    
    // tristate buffers to prevent immediate current requirements from motors
   OBUFT A_1A_tbuf (.I(1'b0), .O(A_1A), .T(!A_1A_o));       // the tristate input is active HIGH, tristate signal being low drives the .I to the motor control chip
   OBUFT A_1B_tbuf (.I(1'b0), .O(A_1B), .T(!A_1B_o));        

   
   OBUFT A_2A_tbuf (.I(1'b0), .O(A_2A), .T(!A_2A_o));
   OBUFT A_2B_tbuf (.I(1'b0), .O(A_2B), .T(!A_2B_o));        
  
    
   OBUFT B_1A_tbuf (.I(1'b0), .O(B_1A), .T(!B_1A_o));
   OBUFT B_1B_tbuf (.I(1'b0), .O(B_1B), .T(!B_1B_o));
    
   OBUFT B_2A_tbuf (.I(1'b0), .O(B_2A), .T(!B_2A_o));
   OBUFT B_2B_tbuf (.I(1'b0), .O(B_2B), .T(!B_2B_o)); 
 
    
    // for simulation purposes, create a commanded heading based on switch inputs
    always@(posedge clk_slow)
    begin
        case (cmd_hdg_sel)
            
            1'b0       :       cmd_hdg <= 24'h006440;      // pi/2 in 14 bit fractional fixed point
            
            1'b1       :       cmd_hdg <= 24'h012D40;      // 3pi/2 in 14 bit fractional fixed point
            
            default     :       cmd_hdg <= 24'h000000;
        
        endcase
    end
    
    
    top_HMC5883L mag_compass (.raw_clk_p(raw_clk_p), .rstn(rstn),
                    // I2C interface
                    .SDA(SDA), .SCL(SCL), 
                    // User/HMC5883L signals
                    .I2C_RST(I2C_RST), .INIT(INIT), .data_out_led(data_out_led), .ack_or_complete(ack_or_complete), .drdy(drdy),
                    .cal_complete(cal_complete), .mag_angle(mag_angle), .mag_angle_valid(mag_angle_valid),
                    // clock signals
                    .clk_fast(clk_fast), .clk_slow(clk_slow), .clk_slow_test());
                    
    top_hdg_ctrl motor_control (.clk(clk_slow), .cal_complete(cal_complete), .motor_enable(/* motor_enable */),
                    .A_1A(A_1A_o), .A_1B(A_1B_o), /* left front motor */ .A_2A(A_2A_o), .A_2B(A_2B_o), /* left rear motor */
                    .B_1A(B_1A_o), .B_1B(B_1B_o), /* right front motor */ .B_2A(B_2A_o), .B_2B(B_2B_o), /* right real motor */
                    .current_hdg(mag_angle), .cmd_hdg(cmd_hdg), .fwd(fwd), .rev(rev));
    
 // DEBUG DEBUG DEBUG DEBUG BELOW.................................................................................................                             
 // debug logic.  UART used to stream mag_angle values to the TeraTerm
/*
 reg [7:0] banner_addr;
 wire banner_done, tx_full;
 wire [7:0] uart_byte;
 reg [1:0] mag_angle_valid_reg;
 wire write;
 
 localparam BANNER_END_ADDR      = 8'h8D;
 localparam ZERO_ADDR            = 8'h00;
 localparam MAG_ANGLE_ADDR       = 8'h7A;

 
 uartlite_verilog #(.C_CLK_FREQ(100000000), .C_BAUDRATE(460800), .C_DATA_BITS(8), .C_USE_PARITY(0),
     .C_ODD_PARITY(0)) debug_uart (.clk(clk_fast), .rst(!rstn), .rx(uart_rx), .tx(uart_tx), .frame_error(), .parity_error(),
     .write(write), .tx_data(uart_byte), .rx_data(), .rx_data_present(), .rx_read(), .tx_buffer_full(tx_full));                                
                                                           
 banner_rom0 banner (.out_byte(uart_byte), .rd_addr(banner_addr), .ANGLE(mag_angle), .cal_complete(cal_complete));
         
 // read address generator for banner_rom0
 always@(posedge clk_fast)
 begin
    
 mag_angle_valid_reg <= {mag_angle_valid_reg [0], mag_angle_valid};
    
 if (!rstn) banner_addr <= ZERO_ADDR;
    else if ((banner_addr < BANNER_END_ADDR) && (!tx_full)) banner_addr <= banner_addr + 1'b1;
        else if (mag_angle_valid_reg [1] && !mag_angle_valid_reg [0]) banner_addr <= MAG_ANGLE_ADDR;    // look for falling edge of mag angle valid
 end
         
 assign banner_done = (banner_addr == BANNER_END_ADDR) ? 1'b1 : 1'b0;
 assign write = (!banner_done && !tx_full);
                        
*/                           
                              
                              
    
endmodule
