`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: Steven Howell
// 
// Create Date:    11:16:07 07/07/2017 
// Design Name: 
// Module Name:    top_HMC5883L 
// Project Name: 
// Target Devices: XC7A35TCSG324-1L
// Tool versions: Vivado 2017.1
// Description: Compass controller/magnetic angle calculator for HMC5883L
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////
module top_HMC5883L (input raw_clk_p, input rstn,
				// I2C interface
				inout SDA, inout SCL, 
				// User / PMOD signals
				input I2C_RST, input INIT, output data_out_led, output ack_or_complete, input drdy,
				input cal_complete, output reg signed [23:0] mag_angle, output reg mag_angle_valid,
				// clock signals
				output clk_fast, output clk_slow, output clk_slow_test);

reg [63:0] debounce_rst, rst_delayed;
reg [15:0] debounce_I2C_RST, debounce_INIT;
reg [1:0] debounce_i2c_rst, debounce_init, oneshot_drdy;
wire clk_i, rst_debounced, valid, complete, ack;
wire [47:0] data_out;
reg [19:0] live_counter;
wire [7:0] slave_addr, reg_addr, reg_addr_init, data_in, init_data;
wire [7:0] slave_addr_f, reg_addr_f, data_in_f;			
reg [7:0] rom_addr;
reg i2c_rst, wr, rd, pulse_i2c_rst, pulse_init, initialize;
reg [1:0] rng_compute;	
wire WR_rom, RD_rom, retry, no_data, no_data_f;
wire [2:0] rd_bytes;
(* FSM_ENCODING="USER" *) reg [2:0] STATE;

wire divide_valid, divide_by_zero;      // wire outputs of divider core.  divide_valid indicates result can be used, divide_by_zero is an error
wire signed [31:0] divide_result, result_leftshift, result_rightshift;       // divide results are fixed point, fractional 15 bit

// HMC5883L high, low, and med value registers for X, Y, and Z channels, dividend and divisor
reg signed [15:0] X_high_reg, Z_high_reg, Y_high_reg, X_low_reg, Z_low_reg, Y_low_reg, dividend, divisor;  // these can be viewed as signed decimal
reg signed [14:0] Xsf, Ysf;             // scale factors--only the fractional portion thereof.  The integer portion assumed to be '1'    

// HMC5883L value wires
wire signed [15:0] X_value, Z_value, Y_value, X_val_corr, Y_val_corr, X_diff, Y_diff;
wire signed [14:0] X_off_u, Y_off_u, X_off_n, Y_off_n, X_off_s, Y_off_s;  // X/Y offsets unscaled (u), X/Y offsets negative corrected (n), and X/Y offsets scaled (s)

wire [23:0] X_atan_in, Y_atan_in;     // ATAN input/output wires, fixed point, fractional 15 bits, 7 bits of padding '0's in most significant bits
wire signed [23:0] atan_result;       // the arctan result is output fixed point, fractional 14 bits
reg signed [23:0] extra_angle;        // extra angle contains the angle necessary to rotate the atan result back to the correct part of the unit circle
wire atan_enable, atan_valid;
reg atan_enable_reg;

localparam	HMC5883L_ADDR	=	8'h3C;			// APPLIES TO HMC5883L, note device address is 7'h1E, shifted left with a '0' LSB indicates write

localparam	CONFIG_A		=	8'd00;          // These are the I2C subaddresses of various registers within the HMC5883L
localparam	CONFIG_B	    =	8'd01;
localparam	MODE_REG     	=	8'd02;
localparam	DATA_X_MSB   	=	8'd03;
localparam	DATA_X_LSB		=	8'd04;
localparam	DATA_Z_MSB		=	8'd05;
localparam	DATA_Z_LSB		=	8'd06;
localparam	DATA_Y_MSB		=	8'd07;
localparam	DATA_Y_LSB		=	8'd08;
localparam	STATUS			=	8'd09;
localparam	ID_A     		=	8'd10;
localparam  ID_B            =   8'd11;
localparam  ID_C            =   8'd12;

localparam	SETUP_5883_START		=	8'h00;				// This is where the 5883 I2C commands start in the i2c_setup_rom addresses
localparam	SETUP_5883_END			=	8'h04;				

localparam  HMC5883_READ            =   8'h05;              // This is where the 5883 read results commands begin in the i2c_setup_rom addresses
localparam  HMC5883_READ_END        =   8'h05;

localparam IDLE 				    =   3'h0;
localparam RUN					    =   3'h1;
localparam ADDR				        =   3'h2;
localparam SETUP 				    =   3'h3;
localparam WAIT_SETUP		        =   3'h4;

localparam RETRY_CNT			    = 10'h3FF;

localparam GENERAL_CALL = 8'h00;

localparam CLEAR = 5'd0;
localparam FULL = 5'd21;

localparam PI_2     =   24'b0000_0000_0110_0100_0100_0000;  // 1.57 in real, fractional 14 bits
localparam PI       =   24'b0000_0000_1100_1000_1100_0000;  // 3.14 in real, fractional 14 bits
localparam tPI_2    =   24'b0000_0001_0010_1101_0100_0000;  // 4.71 in real, fractional 14 bits
localparam tPI      =   24'b0000_0001_1001_0001_1100_0000;  // 6.28 in real, fractional 14 bits

IBUFG clk_ibufg (.I(raw_clk_p), .O(clk_i));				// the raw 100 MHz clock is brought in on IBUFG, sent to MMCM

assign ack_or_complete = ack | complete;                // ack_or_complete indicates I2C transactions happening, just for confidence/debug

always@(posedge clk_fast) debounce_rst <= {debounce_rst [62:0], !rstn};
assign rst_debounced = &debounce_rst;                   // debounce the reset, provide to logic
always@(posedge clk_fast) rst_delayed <= {rst_delayed [62:0], rst_debounced};  // delay the reset to the MMCM.  It will after and be released after the logic reset

// the clock module divides down the 100 MHz board clock with a counter to get the slowest possible clock for I2C and logic use
clock_module_low_power clk_box (.raw_clk(clk_i), .clk_fast(clk_fast), .clk_slow(clk_slow), .rst(rst_delayed[63]));
// raw_clk input at 100MHz for ARTY board

always@(posedge clk_slow)
begin
debounce_I2C_RST <= {debounce_I2C_RST [14:0], I2C_RST};
debounce_INIT <= {debounce_INIT [14:0], INIT};
debounce_init <= {debounce_init[0], &debounce_INIT};
debounce_i2c_rst <= {debounce_i2c_rst[0], &debounce_I2C_RST};
pulse_i2c_rst <= (debounce_i2c_rst == 2'b01)? 1'b1 : 1'b0;
pulse_init <= (debounce_init == 2'b01)? 1'b1 : 1'b0;
live_counter <= live_counter + 1'b1;
oneshot_drdy <= {oneshot_drdy [0], drdy};
end

assign clk_slow_test = live_counter [19];           // create a blinking light to indicate that the slow clock is present in the logic

i2c_interface	#(RETRY_CNT) interface (.clk_slow(clk_slow), .slave_addr(slave_addr_f), .reg_addr(reg_addr_f), .data_in(data_in_f),
									.WR_fast(wr), .RD_fast(rd), .SDA(SDA), .SCL(SCL), .data_out(data_out), .valid(valid),
									.complete(complete), .rst(rst_debounced), .retry(retry), .rd_bytes(rd_bytes), .no_data(no_data_f), .ack(ack));

// assign the X, Y and Z values to data_out vector fields
assign X_value = data_out [47:32];
assign Z_value = data_out [31:16];
assign Y_value = data_out [15:0];

// calculate the difference between the highest X value and lowest X value, highest Y value and lowest Y value for purposes of calibration
assign X_diff = (X_high_reg - X_low_reg);
assign Y_diff = (Y_high_reg - Y_low_reg);

// solve for the Xoffset and Yoffset unscaled values without scale factors -- no division/multiplication required, scale factors assumed to be close to '1'
assign X_off_u = ((X_diff>>>1) - X_high_reg);           
assign Y_off_u = ((Y_diff>>>1) - Y_high_reg); 

// instantiate the multipliers that use scale factors Xsf and Ysf and the X_off_u, Y_off_u to find X_off_s, Y_off_s
// if X_off_u is negative, convert to positive to find the X_off_s value, then add or subtract as required
assign X_off_n = (X_off_u < 0) ? -X_off_u : X_off_u;
assign Y_off_n = (Y_off_u < 0) ? -Y_off_u : Y_off_u;
mult x_sf_mult (.I(X_off_n), .SF(Xsf), .O(X_off_s));        // X_off_s is always a positive value
mult y_sf_mult (.I(Y_off_n), .SF(Ysf), .O(Y_off_s));        // Y_off_s is always a positive value

// find the corrected (calibrated) X and Y values, add the unscaled offset and the scaled portion of the offset, use X_off_s as either negative or positive depending on X/Y_off_u
assign X_val_corr = (X_off_u < 0) ? X_value + X_off_u - X_off_s : X_value + X_off_u + X_off_s;
assign Y_val_corr = (Y_off_u < 0) ? Y_value + Y_off_u - Y_off_s : Y_value + Y_off_u + Y_off_s;

// find unit circle quadrants for calibrated X and Y values.  "Move" to the calculation quadrant by swapping dividend/divisor as required.
// Normalize the dividend/divisor values, and provide to the cordic core to compute ATAN(dividend/divisor).
// "extra_angle" contains the radian value to rotate the ATAN(dividend/divisor) result back to the proper quadrant
always@(posedge clk_slow)
begin
// if cal_complete is valid, then start calculating actual arctan values and compass headings.  Otherwise, should be calculating the Xsf and Ysf
if (cal_complete)
    begin
    // QUADRANT #1 (EAST) 
    if ((X_val_corr > 0) && (((Y_val_corr >= 0) && (X_val_corr > Y_val_corr)) | ((Y_val_corr < 0) && (X_val_corr >= -Y_val_corr)))) 
        begin
        divisor <= X_val_corr;
        dividend <= Y_val_corr;
        extra_angle <= (Y_val_corr < 0) ? tPI : 0;
        end
        // QUADRANT #2 (SOUTH)
        else if ((Y_val_corr < 0) && (((X_val_corr >= 0) && (-Y_val_corr > X_val_corr)) | ((X_val_corr < 0) && (-Y_val_corr >= -X_val_corr))))
            begin
            divisor <= -Y_val_corr;
            dividend <= X_val_corr;
            extra_angle <= tPI_2;
            end
            // QUADRANT #3 (WEST)
            else if ((X_val_corr < 0) && (((Y_val_corr <= 0) && (-X_val_corr > -Y_val_corr)) | ((Y_val_corr > 0) && (-X_val_corr >= Y_val_corr))))
                begin
                divisor <= -X_val_corr;
                dividend <= -Y_val_corr;
                extra_angle <= PI;
                end
                // QUADRANT #4 (NORTH)
                else if ((Y_val_corr > 0) && (((X_val_corr <= 0) && (Y_val_corr > -X_val_corr)) | ((X_val_corr > 0) && (Y_val_corr >= X_val_corr))))
                    begin
                    divisor <= Y_val_corr;
                    dividend <= -X_val_corr;
                    extra_angle <= PI_2;
                    end
                    else divisor <= 0;          // this will result in a divide_by_zero error if above logic does not catch all possibilities
    end // cal_complete is already valid above
    else // the following scale factor calculations happen if cal is not complete
        begin
        dividend <= (X_diff >= Y_diff) ? X_diff : Y_diff;
        divisor <= (X_diff >= Y_diff) ? Y_diff : X_diff;
        if (divide_valid)
            begin
            Ysf <= (X_diff >= Y_diff) ? divide_result [14:0] : 15'h0000;  // if no scale factor, set '0' in the Ysf 
            Xsf <= (X_diff >= Y_diff) ? 15'h0000 : divide_result [14:0];  // if no scale factor, set '0' in the Xsf 
            end
        end                                       
end

// divide the Y value by X, use "divide_by_zero" signal to see any errors in computation of the location in azimuth circle

// rng_compute[1] is used as the tvalid signal going into the divider to delay a single clock cycle from rng[0].  This gives one clock cycle
// to swap the correct values to dividend/divisor in the clock process above
div_gen_0 divider (.aclk(clk_slow), .s_axis_divisor_tvalid(rng_compute[1]), .s_axis_divisor_tdata(divisor), .s_axis_dividend_tvalid(rng_compute[1]),
                            .s_axis_dividend_tdata(dividend), .m_axis_dout_tvalid(divide_valid), .m_axis_dout_tuser(divide_by_zero), 
                            .m_axis_dout_tdata(divide_result));

assign X_atan_in = {7'h00, 17'h08000};                      // this value sets the X to always be '1'  The fractional value from the divider will be the Y value
assign Y_atan_in = {7'h00, result_rightshift [31:15]};      // format of X/Y_atan_in is real, signed fp, fractional 15 bits.  Total field of 17 bits, 7 upper bits are padding


assign result_leftshift = (divide_result <<< 17);           // shift the fraction in the bottom of divide_result to the left, keeping the sign bit
assign result_rightshift = result_leftshift >>> 1;          // adjust the result back to the right one bit (keeping sign) for the format going into the CORDIC


// setup an enable and reset scheme for the ATAN calculation, not using the valid coming from cordic core
always@(posedge clk_slow)
begin
if (rst_debounced) atan_enable_reg <= 1'b0;

    else 
    begin
    
    if (divide_valid) atan_enable_reg <= 1'b1;
    
    else if (atan_valid)
        begin
        mag_angle <= extra_angle + atan_result;  // add/sub ATAN result to extra angle.  format is real, signed fp, fractional 14 bits coming out of CORDIC
        mag_angle_valid <= 1'b1;
            
        atan_enable_reg <= 1'b0;                // test
        end
        else mag_angle_valid <= 1'b0;

    end
end

assign atan_enable = ((divide_valid | atan_enable_reg) && cal_complete);        // if calibration of the HMC5883L isn't complete, don't calculate the arctan values

// compute the arctan of the divided results (1/8th circle, add the missing angles back in later
cordic_0 arctan (.aclk(clk_slow), .aclken(atan_enable), .aresetn(!rst_debounced), .s_axis_cartesian_tvalid(divide_valid),
                                  .s_axis_cartesian_tdata({Y_atan_in, X_atan_in}), .m_axis_dout_tvalid(atan_valid), .m_axis_dout_tdata(atan_result));


always@(posedge clk_slow)
begin

	if (!rstn) 
	   begin
	   STATE <= IDLE;
	   // must initialize the HMC high, low registers for proper operation
	   X_high_reg <= 16'h0000;
	   Z_high_reg <= 16'h0000;
	   Y_high_reg <= 16'h0000;
	   X_low_reg <= 16'hFFFF;
	   Z_low_reg <= 16'hFFFF;
	   Y_low_reg <= 16'hFFFF;
	   end

	   else
		begin
		
		rng_compute [1] <= rng_compute [0];       // delay the enable going to the divider to allow the dividend and divisor to be set properly
		
		case (STATE)
				
				IDLE		:	begin
								if (pulse_init)     // If user presses the INIT pushbutton, run the initialization plus some extra commands
								begin
								STATE <= SETUP;								
								rom_addr <= SETUP_5883_START;     // decide where to enter the command ROM for initialization
								initialize <= 1'b1;
								end
								else if (pulse_i2c_rst)     // If user presses the I2C_RST button, send the global address software reset command
									begin
									STATE <= RUN;
									i2c_rst <= 1'b1;
									wr <= 1'b1;
									end
									else if (oneshot_drdy[1] && !oneshot_drdy[0])
									   begin
									   STATE <= SETUP;								
                                       rom_addr <= HMC5883_READ;     // enter HMC5883L commands to read the results
									   end
									   else
                                           begin
                                           i2c_rst <= 1'b0;
                                           wr <= 1'b0;
                                           initialize <= 1'b0;
                                           rng_compute [0] <= 1'b0;
                                           
                                           // here is where the hMC5883L range and average are computed when rng_compute [0] is high
                                           if (rng_compute [0])
                                           begin
                                           
                                               case (data_out [47])
                                               
                                                    1'b1    :       begin
                                                                    X_low_reg <= (X_value < X_low_reg) ? X_value : X_low_reg;
                                                                    end
                                                                    
                                                    1'b0    :       begin
                                                                    X_high_reg <= (X_value > X_high_reg) ? X_value : X_high_reg;
                                                                    end
                                                                    
                                                    default :       begin
                                                                    X_low_reg <= X_low_reg;
                                                                    X_high_reg <= X_high_reg;
                                                                    end
                                               
                                               endcase
                                               
                                               case (data_out [15])
                                                                                              
                                                    1'b1    :       begin
                                                                    Y_low_reg <= (Y_value < Y_low_reg) ? Y_value : Y_low_reg;
                                                                    end
                                                                                                                   
                                                    1'b0    :       begin
                                                                    Y_high_reg <= (Y_value > Y_high_reg) ? Y_value : Y_high_reg;
                                                                    end
                                                                                                                   
                                                    default :       begin
                                                                    Y_low_reg <= Y_low_reg;
                                                                    Y_high_reg <= Y_high_reg;
                                                                    end
                                                                                              
                                               endcase

                                               case (data_out [31])
                                                                                              
                                                    1'b1    :       begin
                                                                    Z_low_reg <= (Z_value < Z_low_reg) ? Z_value : Z_low_reg;
                                                                    end
                                                                                                                   
                                                    1'b0    :       begin
                                                                    Z_high_reg <= (Z_value > Z_high_reg) ? Z_value : Z_high_reg;
                                                                    end
                                                                                                                   
                                                    default :       begin
                                                                    Z_low_reg <= Z_low_reg;
                                                                    Z_high_reg <= Z_high_reg;
                                                                    end
                                                                                              
                                               endcase                                             
                                           
                                           end
                                           
                                           
                                           end
								end
								
				RUN		    :	begin
								if (complete) STATE <= IDLE;
									else wr <= 1'b0;
								end
								
				SETUP		:	begin
								wr <= WR_rom;
								rd <= RD_rom; 
								STATE <= WAIT_SETUP;
								end
										
				WAIT_SETUP	:	begin
								wr <= 1'b0;
								rd <= 1'b0;
				
								if (complete)
									begin
									if ((rom_addr < SETUP_5883_END) | (!drdy && (rom_addr < HMC5883_READ_END)))
										begin
										rom_addr <= rom_addr + 1'b1;
										STATE <= SETUP;
										end
										else
											begin
											rom_addr <= SETUP_5883_START;
											STATE <= IDLE;
											if (initialize) initialize <= 1'b0;
											    else rng_compute [0] <= 1'b1;
											end
										end
									end	
		
			default		    :	begin
								STATE <= IDLE;
								end
								
			endcase
		
		end
end

assign slave_addr_f = (i2c_rst)? GENERAL_CALL : slave_addr;
assign reg_addr_f = (i2c_rst)? 8'h06 : reg_addr_init;	// 8'h06 = software reset
assign data_in_f = (i2c_rst)? 8'h06 : init_data;
assign no_data_f = (i2c_rst)? 1'b1 : no_data;

assign data_out_led = ^{Z_low_reg, Z_high_reg, divide_by_zero};


i2c_rom #(HMC5883L_ADDR, CONFIG_A, CONFIG_B, MODE_REG, DATA_X_MSB, DATA_X_LSB, DATA_Z_MSB, DATA_Z_LSB, DATA_Y_MSB, DATA_Y_LSB, STATUS, ID_A, ID_B, ID_C) 
		HMC5883L_initialization (.addr(rom_addr), .slave_addr(slave_addr), .reg_addr(reg_addr_init), .data(init_data), .WR(WR_rom), .RD(RD_rom), .retry(retry), .rd_bytes(rd_bytes),
									 .no_data(no_data));




endmodule
